from __future__ import print_function
from collections import Counter
import string
import re
import argparse
import json
import sys
import os

'''KorQuAD v1.0에 대한 공식 평가 스크립트 '''
'''본 스크립트는 SQuAD v1.1 평가 스크립트 https://rajpurkar.github.io/SQuAD-explorer/ 를 바탕으로 작성됨.'''


def normalize_answer(s):
    def remove_(text):
        ''' 불필요한 기호 제거 '''
        text = re.sub("'", " ", text)
        text = re.sub('"', " ", text)
        text = re.sub('《', " ", text)
        text = re.sub('》', " ", text)
        text = re.sub('<', " ", text)
        text = re.sub('>', " ", text)
        text = re.sub('〈', " ", text)
        text = re.sub('〉', " ", text)
        text = re.sub("\(", " ", text)
        text = re.sub("\)", " ", text)
        text = re.sub("‘", " ", text)
        text = re.sub("’", " ", text)
        return text

    def white_space_fix(text):
        return ' '.join(text.split())

    def remove_punc(text):
        exclude = set(string.punctuation)
        return ''.join(ch for ch in text if ch not in exclude)

    def lower(text):
        return text.lower()

    return white_space_fix(remove_punc(lower(remove_(s))))


def f1_score(prediction, ground_truth):
    prediction_tokens = normalize_answer(prediction).split()
    ground_truth_tokens = normalize_answer(ground_truth).split()

    # F1 by character
    prediction_Char = []
    for tok in prediction_tokens:
        now = [a for a in tok]
        prediction_Char.extend(now)

    ground_truth_Char = []
    for tok in ground_truth_tokens:
        now = [a for a in tok]
        ground_truth_Char.extend(now)

    common = Counter(prediction_Char) & Counter(ground_truth_Char)
    num_same = sum(common.values())
    if num_same == 0:
        return 0

    precision = 1.0 * num_same / len(prediction_Char)
    recall = 1.0 * num_same / len(ground_truth_Char)
    f1 = (2 * precision * recall) / (precision + recall)

    return f1


def exact_match_score(prediction, ground_truth):
    return (normalize_answer(prediction) == normalize_answer(ground_truth))


def metric_max_over_ground_truths(metric_fn, prediction, ground_truths):
    scores_for_ground_truths = []
    for ground_truth in ground_truths:
        score = metric_fn(prediction, ground_truth)
        scores_for_ground_truths.append(score)
    return max(scores_for_ground_truths)


def evaluate(dataset, predictions):
    f1 = exact_match = total = 0
    for article in dataset:
        for paragraph in article['paragraphs']:
            for qa in paragraph['qas']:
                total += 1
                if qa['id'] not in predictions:
                    message = 'Unanswered question ' + qa['id'] + \
                              ' will receive score 0.'
                    print(message, file=sys.stderr)
                    continue
                ground_truths = list(map(lambda x: x['text'], qa['answers']))
                prediction = predictions[qa['id']]
                exact_match += metric_max_over_ground_truths(
                    exact_match_score, prediction, ground_truths)
                f1 += metric_max_over_ground_truths(
                    f1_score, prediction, ground_truths)

    exact_match = 100.0 * exact_match / total
    f1 = 100.0 * f1 / total
    return {'exact_match': exact_match, 'f1': f1}


def eval_during_train(args):
    expected_version = 'KorQuAD_v1.0'

    dataset_file = os.path.join(args.data_dir, args.predict_file)
    prediction_file = os.path.join(args.output_dir, 'predictions_.json')

    with open(dataset_file) as dataset_f:
        dataset_json = json.load(dataset_f)
        read_version = "_".join(dataset_json['version'].split("_")[:-1])
        if (read_version != expected_version):
            print('Evaluation expects ' + expected_version +
                  ', but got dataset with ' + read_version,
                  file=sys.stderr)
        dataset = dataset_json['data']
    with open(prediction_file) as prediction_f:
        predictions = json.load(prediction_f)

    return evaluate(dataset, predictions)


if __name__ == '__main__':
    expected_version = 'KorQuAD_v1.0'
    parser = argparse.ArgumentParser(
        description='Evaluation for KorQuAD ' + expected_version)
    parser.add_argument('dataset_file', help='Dataset file')
    parser.add_argument('prediction_file', help='Prediction File')
    args = parser.parse_args()
    with open(args.dataset_file) as dataset_file:
        dataset_json = json.load(dataset_file)
        read_version = "_".join(dataset_json['version'].split("_")[:-1])
        if (read_version != expected_version):
            print('Evaluation expects ' + expected_version +
                  ', but got dataset with ' + read_version,
                  file=sys.stderr)
        dataset = dataset_json['data']
    with open(args.prediction_file) as prediction_file:
        predictions = json.load(prediction_file)
    print(json.dumps(evaluate(dataset, predictions)))


from __future__ import unicode_literals

from builtins import zip

import numpy as np
from sklearn.decomposition import PCA
from matplotlib import pyplot

import matplotlib.pyplot as plt
from sklearn.manifold import TSNE
import gensim

class visualizeW2VPlot:
    def __init__(self):
        name = 'visualizeW2VPlot'

    def load(self, modelFile):
        model = gensim.models.KeyedVectors.load_word2vec_format(modelFile, binary=True, unicode_errors='ignore')
        return model

    def visualizePCA(self, model):
        pyplot.rc('font', family='New Gulim')

        words = ['이재명', '문재인', '승인', '당', '핵', '평화', '정치인', '대표']

        word_vectors = np.vstack([model[w] for w in words])
        twodim = PCA().fit_transform(word_vectors)[:, :2]
        twodim.shape
        plt.figure(figsize=(5, 5))
        plt.scatter(twodim[:, 0], twodim[:, 1], edgecolors='k', c='r')
        for word, (x, y) in zip(words, twodim):
            plt.text(x, y, word)
        plt.axis('off');

        fig1 = plt.gcf()
        plt.show()
        plt.draw()
        fig1.savefig('testPCA.png', dpi=100)


    def visualizeTSNE(self, model, word, vector_size):
        pyplot.rc('font', family='New Gulim')

        arr = np.empty((0, vector_size), dtype='f')
        word_labels = [word]

        # get close words
        close_words = model.similar_by_word(word,topn=20)

        # add the vector for each of the closest words to the array
        arr = np.append(arr, np.array([model[word]]), axis=0)
        for wrd_score in close_words:
            wrd_vector = model[wrd_score[0]]
            word_labels.append(wrd_score[0])
            arr = np.append(arr, np.array([wrd_vector]), axis=0)

        # find tsne coords for 2 dimensions
        tsne = TSNE(n_components=2, random_state=0)
        np.set_printoptions(suppress=True)
        Y = tsne.fit_transform(arr)

        x_coords = Y[:, 0]
        y_coords = Y[:, 1]
        # display scatter plot
        plt.scatter(x_coords, y_coords)

        for label, x, y in zip(word_labels, x_coords, y_coords):
            plt.annotate(label, xy=(x, y), xytext=(0, 0), textcoords='offset points')
        plt.xlim(x_coords.min() + 0.00005, x_coords.max() + 0.00005)
        plt.ylim(y_coords.min() + 0.00005, y_coords.max() + 0.00005)
        plt.show()

if __name__ == '__main__':
    model_file = '../embeddings/word2vec/korean_wiki_w2v.bin'
    model = visualizeW2VPlot().load(model_file)
    mode = 't-sne' # t-sne or pca
    if mode == 'pca':
        visualizeW2VPlot().visualizePCA(model)
    elif mode == 't-sne':
        vector_size = 300
        visualizeW2VPlot().visualizeTSNE(model, '이재명', vector_size)


import pandas as pd
import numpy as np
import torch
from transformers import AutoModel, AutoTokenizer
import os
from dataclasses import dataclass
import torch
from torch import nn

class Config:
    def __init__(self, chatbot_data=None, para_kqc_data = None,
                 questions = None, bert_multilingual = None):
        if chatbot_data is None:
            self.ChatbotData = "../data/chatbot_data.txt"
        else:
            self.ChatbotData = chatbot_data

        if para_kqc_data is None:
            self.ParaKQCData = para_kqc_data
        else:
            self.ParaKQCData = "../data/paraKQC_v1.txt"

        if questions is None:
            self.Questions = questions
        else:
            self.Questions = "../data/questions_embeddings.npy"

        if bert_multilingual is None:
            self.BERTMultiLingual = bert_multilingual
        else:
            self.BERTMultiLingual = "bert-base-multilingual-cased"

        self.Device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')

class Similarity:
    def __init__(self):
        print("Similarity")

    def get_l2_distance(self, x1, x2, dim: int = 1):
        return ((x1 - x2) ** 2).sum(dim=dim) ** .5

    def get_l1_distance(self, x1, x2, dim: int = 1):
        return ((x1 - x2).abs()).sum(dim=dim)

    def get_cosine_similarity(self, x1, x2, dim: int = 1):
        cos = nn.CosineSimilarity(dim=1, eps=1e-6)
        return cos(x1, x2)


class VanillaChatbot:
    def __init__(self, config=None, sim_type: str = 'cos'):
        self.config = config
        print(self.config.BERTMultiLingual)
        self.model = AutoModel.from_pretrained(self.config.BERTMultiLingual)
        self.tokenizer = AutoTokenizer.from_pretrained(self.config.BERTMultiLingual)
        self.questions = self.load_questions(self.config.Questions)
        self.answers = pd.read_csv(self.config.ChatbotData)["A"].tolist()
        self.measure = self.get_similarity_measure(sim_type=sim_type)  # 유사도 측정 객체

    def query(self, question: str, return_answer=False):
        question_tokenized = self.tokenizer(question, return_tensors="pt")
        question_embedded = self.model(**question_tokenized).pooler_output
        similar_question_id = self.get_similar_question_id(question_embedded)
        answer = self.answers[similar_question_id]
        print('챗봇:', answer)
        if return_answer:
            return answer

    def get_similar_question_id(self, q_emb):
        similarities = self.measure(self.questions, q_emb)
        similar_question_id = torch.argmax(similarities).item()
        return similar_question_id

    @staticmethod
    def load_questions(root):
        questions = torch.tensor(np.load(root))
        return questions

    @staticmethod
    def get_similarity_measure(sim_type: str):
        if sim_type == 'cos':
            measure = Similarity().get_cosine_similarity
        elif sim_type == 'l1':
            measure = Similarity().get_l1_distance
        elif sim_type == 'l2':
            measure = Similarity().get_l2_distance

        return measure




# coding=utf-8
# Copyright 2019 SK T-Brain Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import requests
import hashlib

import mxnet as mx
import gluonnlp as nlp
from gluonnlp.model import BERTModel, BERTEncoder

from .utils import download as _download


kobert_models = {
    'mxnet_kobert': {
        'url':
        'https://kobert.blob.core.windows.net/models/kobert/mxnet/mxnet_kobert_45b6957552.params',
        'fname': 'mxnet_kobert_45b6957552.params',
        'chksum': '45b6957552'
    },
    'vocab': {
        'url':
        'https://kobert.blob.core.windows.net/models/kobert/vocab/kobertvocab_f38b8a4d6d.json',
        'fname': 'kobertvocab_f38b8a4d6d.json',
        'chksum': 'f38b8a4d6d'
    }
}


def get_mxnet_kobert_model(use_pooler=True,
                           use_decoder=True,
                           use_classifier=True,
                           ctx=mx.cpu(0),
                           cachedir='./ptr_lm_model'):
    # download model
    model_info = kobert_models['mxnet_kobert']
    model_path = _download(model_info['url'],
                           model_info['fname'],
                           model_info['chksum'],
                           cachedir=cachedir)
    # download vocab
    vocab_info = kobert_models['vocab']
    vocab_path = _download(vocab_info['url'],
                           vocab_info['fname'],
                           vocab_info['chksum'],
                           cachedir=cachedir)
    return get_kobert_model(model_path, vocab_path, use_pooler, use_decoder,
                            use_classifier, ctx)


def get_kobert_model(model_file,
                     vocab_file,
                     use_pooler=True,
                     use_decoder=True,
                     use_classifier=True,
                     ctx=mx.cpu(0)):
    vocab_b_obj = nlp.vocab.BERTVocab.from_json(open(vocab_file, 'rt').read())

    predefined_args = {
        'attention_cell': 'multi_head',
        'num_layers': 12,
        'units': 768,
        'hidden_size': 3072,
        'max_length': 512,
        'num_heads': 12,
        'scaled': True,
        'dropout': 0.1,
        'use_residual': True,
        'embed_size': 768,
        'embed_dropout': 0.1,
        'token_type_vocab_size': 2,
        'word_embed': None,
    }

    encoder = BERTEncoder(attention_cell=predefined_args['attention_cell'],
                          num_layers=predefined_args['num_layers'],
                          units=predefined_args['units'],
                          hidden_size=predefined_args['hidden_size'],
                          max_length=predefined_args['max_length'],
                          num_heads=predefined_args['num_heads'],
                          scaled=predefined_args['scaled'],
                          dropout=predefined_args['dropout'],
                          output_attention=False,
                          output_all_encodings=False,
                          use_residual=predefined_args['use_residual'])

    # BERT
    net = BERTModel(
        encoder,
        len(vocab_b_obj.idx_to_token),
        token_type_vocab_size=predefined_args['token_type_vocab_size'],
        units=predefined_args['units'],
        embed_size=predefined_args['embed_size'],
        embed_dropout=predefined_args['embed_dropout'],
        word_embed=predefined_args['word_embed'],
        use_pooler=use_pooler,
        use_decoder=use_decoder,
        use_classifier=use_classifier)
    net.initialize(ctx=ctx)
    net.load_parameters(model_file, ctx, ignore_extra=True)
    return (net, vocab_b_obj)




#see in glove-win_devc_x64, demo.sh or demo.bat
'''
MacOS and Linux
Go to: https://github.com/stanfordnlp/GloVe

$ git clone http://github.com/stanfordnlp/glove
$ cd glove && make
$ ./demo.sh


Windows 10
https://github.com/anoidgit/GloVe-win

'''



"""
ref: https://github.com/aisolab/nlp_implementation/blob/master/Character-level_Convolutional_Networks_for_Text_Classification/utils.py
"""
import json
import torch
from pathlib import Path
import logging
import os
import random

import numpy as np
import torch
from transformers import BertTokenizer


class Config:
    def __init__(self, json_path):
        with open(json_path, mode='r') as io:
            params = json.loads(io.read())
        self.__dict__.update(params)

    def save(self, json_path):
        with open(json_path, mode='w') as io:
            json.dump(self.__dict__, io, indent=4)

    def update(self, json_path):
        with open(json_path, mode='r') as io:
            params = json.loads(io.read())
        self.__dict__.update(params)

    @property
    def dict(self):
        return self.__dict__


class CheckpointManager:
    def __init__(self, model_dir):
        if not isinstance(model_dir, Path):
            model_dir = Path(model_dir)
        self._model_dir = model_dir

    def save_checkpoint(self, state, filename):
        torch.save(state, self._model_dir / filename)

    def load_checkpoint(self, filename):
        state = torch.load(self._model_dir / filename, map_location=torch.device('cpu'))
        return state


class SummaryManager:
    def __init__(self, model_dir):
        if not isinstance(model_dir, Path):
            model_dir = Path(model_dir)
        self._model_dir = model_dir
        self._summary = {}

    def save(self, filename):
        with open(self._model_dir / filename, mode='w') as io:
            json.dump(self._summary, io, indent=4)

    def load(self, filename):
        with open(self._model_dir / filename, mode='r') as io:
            metric = json.loads(io.read())
        self.update(metric)

    def update(self, summary):
        self._summary.update(summary)

    def reset(self):
        self._summary = {}

    @property
    def summary(self):
        return self._summary




class PadSequence:
    """PadSequence class"""

    def __init__(self, length: int, pad_val: int = 0, clip: bool = True) -> None:
        """Instantiating PadSequence class
        Args:
            length (int): the maximum length to pad/clip the sequence
            pad_val (int): the pad value
            clip (bool): whether to clip the length, if sample length is longer than maximum length
        """
        self._length = length
        self._pad_val = pad_val
        self._clip = clip

    def __call__(self, sample):
        sample_length = len(sample)
        if sample_length >= self._length:
            if self._clip and sample_length > self._length:
                return sample[: self._length]
            else:
                return sample
        else:
            return sample + [self._pad_val for _ in range(self._length - sample_length)]


ADDITIONAL_SPECIAL_TOKENS = ["<e1>", "</e1>", "<e2>", "</e2>"]


def get_label(args):
    return [label.strip() for label in open(os.path.join(args.data_dir, args.label_file), "r", encoding="utf-8")]


def load_tokenizer(args):
    tokenizer = BertTokenizer.from_pretrained(args.model_name_or_path)
    tokenizer.add_special_tokens({"additional_special_tokens": ADDITIONAL_SPECIAL_TOKENS})
    return tokenizer


def write_prediction(args, output_file, preds):
    """
    For official evaluation script
    :param output_file: prediction_file_path (e.g. eval/proposed_answers.txt)
    :param preds: [0,1,0,2,18,...]
    """
    relation_labels = get_label(args)
    with open(output_file, "w", encoding="utf-8") as f:
        for idx, pred in enumerate(preds):
            f.write("{}\t{}\n".format(8001 + idx, relation_labels[pred]))


def init_logger():
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s -   %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        level=logging.INFO,
    )


def set_seed(args):
    random.seed(args.seed)
    np.random.seed(args.seed)
    torch.manual_seed(args.seed)
    if not args.no_cuda and torch.cuda.is_available():
        torch.cuda.manual_seed_all(args.seed)


def compute_metrics(preds, labels):
    assert len(preds) == len(labels)
    return acc_and_f1(preds, labels)


def simple_accuracy(preds, labels):
    return (preds == labels).mean()


def acc_and_f1(preds, labels, average="macro"):
    acc = simple_accuracy(preds, labels)
    return {
        "acc": acc,
        "f1": official_f1(),
    }


EVAL_DIR = "../eval"


def official_f1():
    # Run the perl script
    try:
        cmd = "perl {0}/semeval2010_task8_scorer-v1.2.pl {0}/proposed_answers.txt {0}/answer_keys.txt > {0}/result.txt".format(
            EVAL_DIR
        )
        os.system(cmd)
    except:
        raise Exception("perl is not installed or proposed_answers.txt is missing")

    with open(os.path.join(EVAL_DIR, "result.txt"), "r", encoding="utf-8") as f:
        macro_result = list(f)[-1]
        macro_result = macro_result.split(":")[1].replace(">>>", "").strip()
        macro_result = macro_result.split("=")[1].strip().replace("%", "")
        macro_result = float(macro_result) / 100

    return macro_result


from functools import reduce

from torch.nn import ModuleList
import torch.nn.functional as F
import copy
import math
import torch
import numpy as np
"""
ModuleList는 목록에 하위 모듈을 보관하것
이때 모듈들은 파이썬 리스트들 처럼 인덱스를 사용할 수 있다.
"""
def clones(module, N):
  return ModuleList([copy.deepcopy(module) for i in range(N)])

"""
디코더에서 어텐션 스코어 매트릭스에서
이후의 값들에 대해 -∞으로 마스킹 처리해주기 위한 함수
(1, size, size)의 마스크를 리턴한다.
"""
def subsequent_mask(size):
  "Mask out subsequent positions."
  attn_shape = (1, size, size)
  subsequent_mask = np.triu(np.ones(attn_shape), k=1).astype('uint8')
  return torch.from_numpy(subsequent_mask) == 0

def log(t, eps=1e-9):
    return torch.log(t + eps)

def gumbel_noise(t):
    noise = torch.zeros_like(t).uniform_(0, 1)
    return -log(-log(noise))

def gumbel_sample(t, temperature = 1.):
    return ((t / temperature) + gumbel_noise(t)).argmax(dim=-1)

def prob_mask_like(t, prob):
    return torch.zeros_like(t).float().uniform_(0, 1) < prob

def mask_with_tokens(t, token_ids):
    init_no_mask = torch.full_like(t, False, dtype=torch.bool)
    mask = reduce(lambda acc, el: acc | (t == el), token_ids, init_no_mask)
    return mask

def get_mask_subset_with_prob(mask, prob):
    batch, seq_len, device = *mask.shape, mask.device
    max_masked = math.ceil(prob * seq_len)

    num_tokens = mask.sum(dim=-1, keepdim=True)
    mask_excess = (mask.cumsum(dim=-1) > (num_tokens * prob).ceil())
    mask_excess = mask_excess[:, :max_masked]

    rand = torch.rand((batch, seq_len), device=device).masked_fill(~mask, -1e9)
    _, sampled_indices = rand.topk(max_masked, dim=-1)
    sampled_indices = (sampled_indices + 1).masked_fill_(mask_excess, 0)

    new_mask = torch.zeros((batch, seq_len + 1), device=device)
    new_mask.scatter_(-1, sampled_indices, 1)
    return new_mask[:, 1:].bool()

def temperature_sampling(logits, temperature):
  if temperature is None or temperature == 0.0:
    return torch.argmax(logits)
  probs = F.softmax(logits / temperature)
  pred_ids = probs.cpu().multinomial(probs.size()[1], replacement=False)
  return pred_ids


# coding=utf-8
# Copyright 2019 SK T-Brain Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import requests
import hashlib

kobert_models = {
    'onnx_kobert': {
        'url':
        'https://kobert.blob.core.windows.net/models/kobert/onnx/onnx_kobert_44529811f0.onnx',
        'fname': 'onnx_kobert_44529811f0.onnx',
        'chksum': '44529811f0'
    },
    'tokenizer': {
        'url':
        'https://kobert.blob.core.windows.net/models/kobert/tokenizer/tokenizer_78b3253a26.model',
        'fname': 'tokenizer_78b3253a26.model',
        'chksum': '78b3253a26'
    }
}


def download(url, filename, chksum, cachedir='./ptr_lm_model'):
    f_cachedir = os.path.expanduser(cachedir)
    os.makedirs(f_cachedir, exist_ok=True)
    file_path = os.path.join(f_cachedir, filename)
    if os.path.isfile(file_path):
        if hashlib.md5(open(file_path,
                            'rb').read()).hexdigest()[:10] == chksum:
            print('using cached model')
            return file_path
    with open(file_path, 'wb') as f:
        response = requests.get(url, stream=True)
        total = response.headers.get('content-length')

        if total is None:
            f.write(response.content)
        else:
            downloaded = 0
            total = int(total)
            for data in response.iter_content(
                    chunk_size=max(int(total / 1000), 1024 * 1024)):
                downloaded += len(data)
                f.write(data)
                done = int(50 * downloaded / total)
                sys.stdout.write('\r[{}{}]'.format('█' * done,
                                                   '.' * (50 - done)))
                sys.stdout.flush()
    sys.stdout.write('\n')
    assert chksum == hashlib.md5(open(
        file_path, 'rb').read()).hexdigest()[:10], 'corrupted file!'
    return file_path


def get_onnx(cachedir='./ptr_lm_model'):
    """Get KoBERT ONNX file path after downloading
    """
    model_info = kobert_models['onnx_kobert']
    return download(model_info['url'],
                    model_info['fname'],
                    model_info['chksum'],
                    cachedir=cachedir)

def get_tokenizer(cachedir='./ptr_lm_model'):
    """Get KoBERT Tokenizer file path after downloading
    """
    model_info = kobert_models['tokenizer']
    return download(model_info['url'],
                        model_info['fname'],
                        model_info['chksum'],
                        cachedir=cachedir)    




from yonlu.word_embeddings.word_embeddings import Word2Vec

word2vec = Word2Vec()
mode = 'unfiltered'
mecab_path = 'C:\\mecab\\mecab-ko-dic'
stopword_file = '../stopwords/stopwordsKor.txt'
files = []
files.append('../data/content.txt')
is_directory=False
doc_index=-1
max=-1
is_mecab=False
word2vec.preprocessing(mode,is_mecab,mecab_path,stopword_file,files,is_directory,doc_index,max)

min_count=1
window=5
size=200
negative=0
word2vec.train(min_count, window, size, negative)

model_file = 'word2vec_1.bin'
binary=True;
word2vec.save_model(model_file, binary)





import os
import openai

openai.api_key = "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"
completion = openai.Completion()

start_chat_log = '''Human: Hello, who are you?
AI: I am doing great. How can I help you today?
'''

def ask(question, chat_log=None):
    if chat_log is None:
        chat_log = start_chat_log
    prompt = f'{chat_log}Human: {question}\nAI:'
    response = completion.create(
        prompt=prompt, engine="davinci", stop=['\nHuman'], temperature=0.0,
        top_p=1, frequency_penalty=0, presence_penalty=0.6, best_of=1,
        max_tokens=150)
    answer = response.choices[0].text.strip()
    return answer

question = 'Who played Forrest Gump in the movie?'
answer = ask(question)
print(answer)

def append_interaction_to_chat_log(question, answer, chat_log=None):
    if chat_log is None:
        chat_log = start_chat_log
    return f'{chat_log}Human: {question}\nAI: {answer}\n'

chat_log = None
answer = ask(question, chat_log)
print(answer)

chat_log = append_interaction_to_chat_log(question, answer, chat_log)
question = 'Was he in any other great roles?'
answer = ask(question, chat_log)

print(answer)


import torch
from transformers import GPT2LMHeadModel, GPT2Tokenizer

# initialize tokenizer and model from pretrained GPT2 model
tokenizer = GPT2Tokenizer.from_pretrained('gpt2')
model = GPT2LMHeadModel.from_pretrained('gpt2')

sequence = "Worldwide, numerous memorials have been dedicated to Churchill. His statue in Parliament Square was unveiled by his widow Clementine in 1973 and is one of only twelve in the square, all of prominent political figures, including Churchill's friend Lloyd George and his India policy nemesis Gandhi.[459][460] Elsewhere in London, the wartime Cabinet War Rooms have been renamed the Churchill Museum and Cabinet War Rooms.[461] Churchill College, Cambridge, was established as a national memorial to Churchill. An indication of Churchill's high esteem in the UK is the result of the 2002 BBC poll, attracting 447,423 votes, in which he was voted the greatest Briton of all time, his nearest rival being Isambard Kingdom Brunel some 56,000 votes behind.[462] He is one of only eight people to be granted honorary citizenship of the United States; others include Lafayette, Raoul Wallenberg and Mother Teresa.[463] The United States Navy honoured him in 1999 by naming a new Arleigh Burke-class destroyer as the USS Winston S. Crchill.[464] Other memorials in North America include the National Churchill Museum in Fulton, Missouri, where he made the 1946 \"Iron Curtain\" speech; Churchill Square in central Edmonton, Alberta; and the Winston Churchill Range, a mountain range northwest of Lake Louise, also in Alberta, which was renamed after Churchill in 1956.[465]"

#Because we are using PyTorch, we add return_tensor='pt',
# if using TensorFlow, we would use return_tensor='tf'.
inputs = tokenizer.encode(sequence, return_tensors='pt')

# To generate text with GPT-2, all we do is call the model.generate method
# we pass a maximum output length of 200 tokens
outputs = model.generate(inputs, max_length=200, do_sample=True)

# Our generate step outputs an array of tokens rather than words.
# To convert these tokens into words, we need to .decode them
decoded = tokenizer.decode(outputs[0], skip_special_tokens=True)
print(decoded)



from yonlu.convAI import bert_mini_chatbot
import fire

from yonlu.convAI.bert_mini_chatbot import Config
from yonlu.convAI.export_question_embeddings import export_question_embeddings
from os.path import exists

EXIT = "대화종료"

if __name__ == "__main__":
    chatbotData = "../data/ChatbotData.csv"
    paraKQCData = "../data/paraKQC_v1.txt"
    questions = "../data/questions_embeddings.npy"
    BERTMultiLingual = "bert-base-multilingual-cased"
    config = Config(chatbot_data=chatbotData, para_kqc_data=paraKQCData,
                    questions=questions, bert_multilingual=BERTMultiLingual)

    if exists(questions) != True:
        fire.Fire({"run": export_question_embeddings(config)})

    print('*' * 150)
    print('환영합니다! 간단한 챗봇을 사용해보세요')
    print('*' * 150)


    bot = bert_mini_chatbot.VanillaChatbot(config)
    while True:
        text = input("할 말을 입력해주세요(종료시 '대화종료' 입력): ")
        if text == EXIT:
            break
        bot.query(text)
    print('*' * 150)
    print('챗봇을 종료합니다. 다음에 또 만나요!')
    print('*' * 150)


from yonlu.word_embeddings.word_embeddings import FastText

fasttext = FastText()
binary=True
model_file = 'fasttext.bin'
fasttext.load_model(model_file)
mode = 'jamo_split_filtered'
print(fasttext.most_similar(mode, positives=['이재명', '경제'], negatives=['정치인'], topn=10))
#print(fasttext.most_similar(mode, positives=['이재명'], negatives=[], topn=10))

print('-----------------------------------')

print(fasttext.similar_by_word(mode, '이재명'))


"""
Each entry in personachat is a dict with two keys personality and utterances, the dataset is a list of entries.
personality:  list of strings containing the personality of the agent
utterances: list of dictionaries, each of which has two keys which are lists of strings.
    candidates: [next_utterance_candidate_1, ..., next_utterance_candidate_19]
        The last candidate is the ground truth response observed in the conversational data
    history: [dialog_turn_0, ... dialog_turn N], where N is an odd number since the other user starts every conversation.
Preprocessing:
    - Spaces before periods at end of sentences
    - everything lowercase
"""
EXAMPLE_ENTRY = {
    "personality": ["i like to remodel homes .", "i like to go hunting .",
                    "i like to shoot a bow .", "my favorite holiday is halloween ."],
    "utterances": [
        {"candidates": [
            "my mom was single with 3 boys , so we never left the projects .",
            "i try to wear all black every day . it makes me feel comfortable .",
            "well nursing stresses you out so i wish luck with sister",
            "yeah just want to pick up nba nfl getting old",
            "i really like celine dion . what about you ?", "no . i live near farms .",
            "i wish i had a daughter , i'm a boy mom . they're beautiful boys though still "
            "lucky",
            "yeah when i get bored i play gone with the wind my favorite movie .",
            "hi how are you ? i'm eating dinner with my hubby and 2 kids .",
            "were you married to your high school sweetheart ? i was .",
            "that is great to hear ! are you a competitive rider ?",
            "hi , i'm doing ok . i'm a banker . how about you ?", "i'm 5 years old",
            "hi there . how are you today ?",
            "i totally understand how stressful that can be .",
            "yeah sometimes you do not know what you are actually watching",
            "mother taught me to cook ! we are looking for an exterminator .",
            "i enjoy romantic movie . what is your favorite season ? mine is summer .",
            "editing photos takes a lot of work .",
            "you must be very fast . hunting is one of my favorite hobbies ."
        ],
        "history": [
            "hi , how are you doing ? i'm getting ready to do some cheetah chasing to stay in shape .",
        ]
        },
        {"candidates": ["hello i am doing well how are you ?",
                        "ll something like that . do you play games ?",
                        "does anything give you relief ? i hate taking medicine "
                        "for mine .",
                        "i decorate cakes at a local bakery ! and you ?",
                        "do you eat lots of meat",
                        "i am so weird that i like to collect people and cats",
                        "how are your typing skills ?",
                        "yeah . i am headed to the gym in a bit to weight lift .",
                        "yeah you have plenty of time",
                        "metal is my favorite , but i can accept that people "
                        "listen to country . haha",
                        "that's why you desire to be controlled . let me control "
                        "you person one .",
                        "two dogs they are the best , how about you ?",
                        "you do art ? what kind of art do you do ?",
                        "i love watching baseball outdoors on sunny days .",
                        "oh i see . do you ever think about moving ? i do , "
                        "it is what i want .",
                        "sure . i wish it were winter . the sun really hurts my "
                        "blue eyes .",
                        "are we pretending to play tennis",
                        "i am rich and have all of my dreams fulfilled already",
                        "they tire me so , i probably sleep about 10 hrs a day "
                        "because of them .",
                        "i also remodel homes when i am not out bow hunting ."],
         "history": [
            "hi , how are you doing ? i'm getting ready to do some cheetah chasing to stay in shape .",
             "you must be very fast . hunting is one of my favorite hobbies .",
             "i am ! for my hobby i like to do canning or some whittling .",]
         },
        {"candidates": ["yes they do but i say no to them lol",
                        "i have trouble getting along with family .",
                        "i live in texas , what kind of stuff do you do in "
                        "toronto ?",
                        "that's so unique ! veganism and line dancing usually "
                        "don't mix !",
                        "no , it isn't that big . do you travel a lot",
                        "that's because they are real ; what do you do for "
                        "work ?",
                        "i am lazy all day lol . my mom wants me to get a job "
                        "and move out",
                        "i was born on arbor day , so plant a tree in my name",
                        "okay , i should not tell you , its against the rules "
                        "but my name is sarah , call me o",
                        "hello how are u tonight",
                        "cool . . . my parents love country music that's why i "
                        "hate it",
                        "i am an accountant . what do you do ?",
                        "what do your parents do ? my dad is a mechanic .",
                        "how are you liking it ?",
                        "i really am too . great talking to you too .",
                        "cool . whats it like working there ?",
                        "one daughter . she's pre med",
                        "no and all men is taller than me why can't i find a "
                        "man to dance with",
                        "i live in utah , and my family live in england , "
                        "so i understand",
                        "that's awesome . do you have a favorite season or "
                        "time of year ?"],
         "history": [
             "hi , how are you doing ? i'm getting ready to do some cheetah chasing to stay in shape .",
             "you must be very fast . hunting is one of my favorite hobbies .",
             "i am ! for my hobby i like to do canning or some whittling .",
             "i also remodel homes when i am not out bow hunting .",
             "that's neat . when i was in high school i placed 6th in 100m dash !", ]
         },
    ]
}



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import torch
from transformers import BertForQuestionAnswering
from transformers import BertTokenizer

coqa = pd.read_json('../data/coqa-train-v1.0.json')
print(coqa.head())

#data cleaning
del coqa["version"]
cols = ["text","question","answer"]

comp_list = []
for index, row in coqa.iterrows():
    for i in range(len(row["data"]["questions"])):
        temp_list = []
        temp_list.append(row["data"]["story"])
        temp_list.append(row["data"]["questions"][i]["input_text"])
        temp_list.append(row["data"]["answers"][i]["input_text"])
        comp_list.append(temp_list)

new_df = pd.DataFrame(comp_list, columns=cols)

new_df.to_csv("../data/CoQA_data.csv", index=False)
data = pd.read_csv("../data/CoQA_data.csv")
print(data.head())

print("Number of question and answers: ", len(data))

#Building the QA engine
model = BertForQuestionAnswering.from_pretrained('bert-large-uncased-whole-word-masking-finetuned-squad')
tokenizer = BertTokenizer.from_pretrained('bert-large-uncased-whole-word-masking-finetuned-squad')

random_num = np.random.randint(0,len(data))
question = data["question"][random_num]
text = data["text"][random_num]

print(question, "\n", text)

input_ids = tokenizer.encode(question, text)
print("The input has a total of {} tokens.".format(len(input_ids)))

tokens = tokenizer.convert_ids_to_tokens(input_ids)

for token, id in zip(tokens, input_ids):
    print('{:8}{:8,}'.format(token,id))

#first occurence of [SEP] token
sep_idx = input_ids.index(tokenizer.sep_token_id)
print(sep_idx)

#number of tokens in segment A - question
num_seg_a = sep_idx+1
print(num_seg_a)

#number of tokens in segment B - text
num_seg_b = len(input_ids) - num_seg_a
print(num_seg_b)

segment_ids = [0]*num_seg_a + [1]*num_seg_b
print(segment_ids)

assert len(segment_ids) == len(input_ids)

#token input_ids to represent the input
#token segment_ids to differentiate segments - text and question
output = model(torch.tensor([input_ids]), token_type_ids=torch.tensor([segment_ids]))

#tokens with highest start and end scores
answer_start = torch.argmax(output.start_logits)
answer_end = torch.argmax(output.end_logits)

if answer_end >= answer_start:
    answer = " ".join(tokens[answer_start:answer_end + 1])
else:
    print("I am unable to find the answer to this question. Can you please ask another question?")

print("Text:\n{}".format(text.capitalize()))
print("\nQuestion:\n{}".format(question.capitalize()))
print("\nAnswer:\n{}.".format(answer.capitalize()))

#Visualization of the tokens
start_scores = output.start_logits.detach().numpy().flatten()
end_scores = output.end_logits.detach().numpy().flatten()

token_labels = []
for i, token in enumerate(tokens):
    token_labels.append("{}-{}".format(token,i))

print(len(token_labels))
#first 100 tokens
plt.rcParams["figure.figsize"] = (20,10)
ax = sns.barplot(x=token_labels[:80], y=start_scores[:80], ci=None)
ax.set_xticklabels(ax.get_xticklabels(), rotation=90, ha="center")
ax.grid(True)
plt.title("Start word scores")
plt.show()

#last 100 tokens
plt.rcParams["figure.figsize"] = (20,10)
ax = sns.barplot(x=token_labels[-80:], y=start_scores[-80:], ci=None)
ax.set_xticklabels(ax.get_xticklabels(), rotation=90, ha="center")
ax.grid(True)
plt.title("Start word scores")
plt.show()

#first 100 tokens
plt.rcParams["figure.figsize"] = (20,10)
ax = sns.barplot(x=token_labels[:80], y=end_scores[:80], ci=None)
ax.set_xticklabels(ax.get_xticklabels(), rotation=90, ha="center")
ax.grid(True)
plt.title("End word scores")
plt.show()

#last 100 tokens
plt.rcParams["figure.figsize"] = (20,10)
ax = sns.barplot(x=token_labels[-80:], y=end_scores[-80:], ci=None)
ax.set_xticklabels(ax.get_xticklabels(), rotation=90, ha="center")
ax.grid(True)
plt.title("End word scores")
plt.show()

#Code to join the broken words
answer = tokens[answer_start]
for i in range(answer_start+1, answer_end+1):
    if tokens[i][0:2] == "##":
        answer += tokens[i][2:]
    else:
        answer += " " + tokens[i]


def question_answer(question, text):
    # tokenize question and text in ids as a pair
    input_ids = tokenizer.encode(question, text)

    # string version of tokenized ids
    tokens = tokenizer.convert_ids_to_tokens(input_ids)

    # segment IDs
    # first occurence of [SEP] token
    sep_idx = input_ids.index(tokenizer.sep_token_id)

    # number of tokens in segment A - question
    num_seg_a = sep_idx + 1

    # number of tokens in segment B - text
    num_seg_b = len(input_ids) - num_seg_a

    # list of 0s and 1s
    segment_ids = [0] * num_seg_a + [1] * num_seg_b

    assert len(segment_ids) == len(input_ids)

    # model output using input_ids and segment_ids
    output = model(torch.tensor([input_ids]), token_type_ids=torch.tensor([segment_ids]))

    # reconstructing the answer
    answer_start = torch.argmax(output.start_logits)
    answer_end = torch.argmax(output.end_logits)

    if answer_end >= answer_start:
        answer = tokens[answer_start]
        for i in range(answer_start + 1, answer_end + 1):
            if tokens[i][0:2] == "##":
                answer += tokens[i][2:]
            else:
                answer += " " + tokens[i]

    if answer.startswith("[CLS]"):
        answer = "Unable to find the answer to your question."

    #     print("Text:\n{}".format(text.capitalize()))
    #     print("\nQuestion:\n{}".format(question.capitalize()))
    print("\nAnswer:\n{}".format(answer.capitalize()))

text = """New York (CNN) -- More than 80 Michael Jackson collectibles -- including the late pop star's famous rhinestone-studded glove from a 1983 performance -- were auctioned off Saturday, reaping a total $2 million. Profits from the auction at the Hard Rock Cafe in New York's Times Square crushed pre-sale expectations of only $120,000 in sales. The highly prized memorabilia, which included items spanning the many stages of Jackson's career, came from more than 30 fans, associates and family members, who contacted Julien's Auctions to sell their gifts and mementos of the singer. Jackson's flashy glove was the big-ticket item of the night, fetching $420,000 from a buyer in Hong Kong, China. Jackson wore the glove at a 1983 performance during \"Motown 25,\" an NBC special where he debuted his revolutionary moonwalk. Fellow Motown star Walter \"Clyde\" Orange of the Commodores, who also performed in the special 26 years ago, said he asked for Jackson's autograph at the time, but Jackson gave him the glove instead. "The legacy that [Jackson] left behind is bigger than life for me,\" Orange said. \"I hope that through that glove people can see what he was trying to say in his music and what he said in his music.\" Orange said he plans to give a portion of the proceeds to charity. Hoffman Ma, who bought the glove on behalf of Ponte 16 Resort in Macau, paid a 25 percent buyer's premium, which was tacked onto all final sales over $50,000. Winners of items less than $50,000 paid a 20 percent premium."""
question = "Where was the Auction held?"

question_answer(question, text)

print("Original answer:\n", data.loc[data["question"] == question]["answer"].values[0])





from pathlib import Path
import shutil
import unittest

from transformers import OpenAIGPTTokenizer, GPT2Tokenizer
from chatbot_trainer import ATTR_TO_SPECIAL_TOKEN, SPECIAL_TOKENS

class TestSpecialTokenTreatment(unittest.TestCase):

    def setUp(self):
        self.save_dir = Path('utest_save_dir')
        self.save_dir.mkdir(exist_ok=True)

    def tearDown(self):
        shutil.rmtree(self.save_dir)

    def test_special_tokens_checkpoint_behavior(self):
        toks = [OpenAIGPTTokenizer.from_pretrained('openai-gpt'), GPT2Tokenizer.from_pretrained('gpt2')]
        for tok in toks:
            self.assertEqual(len(tok.added_tokens_encoder), 0)
            tok.add_special_tokens(ATTR_TO_SPECIAL_TOKEN)
            self.assertEqual(len(tok.added_tokens_encoder), 5)
            # Make sure we never split
            self.assertEqual(len(tok.tokenize("<bos> <speaker1>")), 2)
            ids = tok.convert_tokens_to_ids(SPECIAL_TOKENS)
            self.assertTrue(all([x > 0 for x in ids]),
                            f'some tokens failed to tokenize {SPECIAL_TOKENS} -> {ids}')
            # Need to mantain indices through save. (this is also tested in pytorch-transformers)
            tok.save_pretrained(self.save_dir)
            tok_loaded = tok.from_pretrained(str(self.save_dir))
            ids2 = tok_loaded.convert_tokens_to_ids(SPECIAL_TOKENS)
            self.assertListEqual(ids, ids2)



from __future__ import print_function
import sys
import os
import argparse


from urllib.request import urlopen
from urllib.parse import urlencode, urlparse
from urllib.error import URLError, HTTPError


from lxml import html, etree
import json
import time


def url_fetch(targeturl, attempt=0, num_retries=5, retry_interval=10):
    """Fetch a

    Args:
        targeturl (str): URL to fetch
        attempt (int): Current attempt number
        num_retries (int): Number of times to retry before throwing an exception
        retry_interval (int, optional): Amount of time to wait between retries in seconds

    Returns:
        TYPE: Description
    """
    try:
        img_data = urlopen(targeturl).read()
        return img_data
    except URLError as ue:
        raise ue


def download_and_save_image(imgurl, save_dir, num_retries=5, retry_interval=10):
    """Download and save an image. The save name is dervied from the url

    Args:
        imgurl (str): Image URL
        save_dir (str): Output directory
        num_retries (int, optional): Number of times to retry before throwing an exception
        retry_interval (int, optional): Amount of time to wait between retries in seconds

    Returns:
        dict: Returns dict containing the path and the id of image saved
              ```
              {
                    "path": str, // image save path,
                    "img_id": str, // image id
              }
              ```
    """
    parse_result = urlparse(imgurl)
    img_name = os.path.basename(parse_result.path)
    img_id = img_name.split(".")[0]
    img_data = url_fetch(imgurl, attempt=0, num_retries=num_retries, retry_interval=retry_interval)
    save_name = os.path.join(save_dir, img_name)
    with open(save_name, "wb") as f:
        f.write(img_data)
    return {"path": save_name, "img_id": img_id}


def normalize_search_term(search_term):
    # convert a search query into proper folder name
    return search_term.replace(" ", "_")


def create_save_dir_name(search_term, out_dir):
    """Create a directory name for saving specific search results

    Args:
        search_term (str): Search query
        out_dir (str): Output directory

    Returns:
        str: Directory in which to store images
    """
    dirname = normalize_search_term(search_term)
    # c = 0
    # while os.path.exists(os.path.join(out_dir, dirname)):
    #     dirname += "(%d)"%c
    #     c+=1
    return os.path.join(out_dir, dirname)

def tags(search_term, out_dir):
    """Fetch and save icons for given search term

    Args:
        search_term (str):search query
        out_dir (str): directory in which to save images

    Returns:
        list: List of dicts, dict is as returned by `download_and_save_image`
    """
    search_url = "https://www.flaticon.com/search?" + urlencode({"word": search_term})
    # download and parse html
    try:
        page = urlopen(search_url)
    except HTTPError as he:
        if he.code == 404:
            print("No search results for {}, skipping.".format(search_term))
            with open(out_dir + "_failed.txt", "w") as f:
                f.write(he.reason)
            return []

def fetch_icons(search_term, out_dir):
    """Fetch and save icons for given search term

    Args:
        search_term (str):search query
        out_dir (str): directory in which to save images

    Returns:
        list: List of dicts, dict is as returned by `download_and_save_image`
    """
    search_url = "https://www.flaticon.com/search?" + urlencode({"word": search_term})
    # download and parse html
    try:
        page = urlopen(search_url)
    except HTTPError as he:
        if he.code == 404:
            print("No search results for {}, skipping.".format(search_term))
            with open(out_dir + "_failed.txt", "w") as f:
                f.write(he.reason)
            return []
    src = page.read().decode("UTF-8")
    htmlpage = html.fromstring(src)
    icon_paths = '//div[@class="icon--holder"]/img'
    image_tags = htmlpage.xpath(icon_paths)
    # extract urls
    image_urls = [tag.get("src") for tag in image_tags]
    results = []
    # download and save
    for iurl in image_urls:
        result = download_and_save_image(iurl, out_dir)
        print("[%s]: saved" % search_term, result["img_id"], "at", result["path"])
        results.append(result)
    return results


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("--inp-file", help="File containing search terms", required=True)
    parser.add_argument("--out-dir", help="Output directory", default="downloaded_images")
    args = parser.parse_args()

    inp_file = args.inp_file
    out_dir = args.out_dir

    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    with open(inp_file, "r") as f:
        for line in f:
            print("fetching for:", line)
            search_term = line.strip(" \n \t, ; ").lower()
            if len(search_term) == 0:
                continue
            # create a valid directory name
            st_out_dir = create_save_dir_name(search_term, out_dir)
            if not os.path.exists(st_out_dir):
                os.makedirs(st_out_dir)
            # fetch icons
            results = fetch_icons(search_term, st_out_dir)


import gensim
import networkx as nx
from gensim.models import KeyedVectors
from node2vec import Node2Vec
import operator

# Embed edges using Hadamard method
from node2vec.edges import HadamardEmbedder
import multiprocessing

class Node2VecModel:
    def __init__(self):
        self.model = None
        self.G = nx.Graph()

    def create_random_graph(self):
        # Create a graph
        self.G = nx.fast_gnp_random_graph(n=100, p=0.5)

    def create_graph(self, co_occurrence, word_hist, threshold):
        filtered_word_list = []
        for pair in co_occurrence:
            node1 = ''
            node2 = ''
            for inner_pair in pair:
                if type(inner_pair) is tuple:
                    node1 = inner_pair[0]
                    node2 = inner_pair[1]
                elif type(inner_pair) is str:
                    inner_pair = inner_pair.split()
                    if len(inner_pair) == 2:
                        node1 = inner_pair[0]
                        node2 = inner_pair[1]
                elif type(inner_pair) is int:
                    if float(inner_pair) >= threshold:
                        # print ("X " + node1 + " == " + node2 + " == " + str(inner_pair) + " : " + str(tuple[node1]))
                        self.G.add_edge(node1, node2, weight=float(inner_pair))
                        if node1 not in filtered_word_list:
                            filtered_word_list.append(node1)
                        if node2 not in filtered_word_list:
                            filtered_word_list.append(node2)
                elif type(inner_pair) is float:
                    if float(inner_pair) >= threshold:
                        # print ("X " + node1 + " == " + node2 + " == " + str(inner_pair) + " : ")
                        self.G.add_edge(node1, node2, weight=float(inner_pair))
                        if node1 not in filtered_word_list:
                            filtered_word_list.append(node1)
                        if node2 not in filtered_word_list:
                            filtered_word_list.append(node2)

        for word in word_hist:
            if str(word) in filtered_word_list:
                self.G.add_node(word, count=word_hist[word])

        print(self.G.number_of_nodes())

    def train(self, dimensions, walk_length, num_walks, min_count=1):
        cores = multiprocessing.cpu_count()  # Count the number of cores in a computer
        # Precompute probabilities and generate walks
        node2vec = Node2Vec(self.G,
                            dimensions=dimensions,
                            walk_length=walk_length,
                            num_walks=num_walks,
                            workers=cores-1)

        ## if d_graph is big enough to fit in the memory, pass temp_folder which has enough disk space
        # Note: It will trigger "sharedmem" in Parallel, which will be slow on smaller graphs
        # node2vec = Node2Vec(graph, dimensions=64, walk_length=30, num_walks=200, workers=4, temp_folder="/mnt/tmp_data")

        # Embed
        self.model = node2vec.fit(window=10, min_count=min_count,
                                  batch_words=4)  # Any keywords acceptable by gensim.Word2Vec can be passed, `diemnsions` and `workers` are automatically passed (from the Node2Vec constructor)

    def save_graph(self, graph_file):
        nx.write_graphml_lxml(self.G, graph_file)

    def load_graph(self, graph_file):
        self.G = nx.read_graphml(graph_file)

    def compute_centrality(self, top_n=30):
        print('network statistics')
        print(nx.info(self.G))

        degree = nx.degree_centrality(self.G)
        degree_l = sorted(degree.items(), key=operator.itemgetter(1), reverse=True)
        print('degree centrality\n')
        for i, n in enumerate(degree_l,100):
            print(str(n))
        print()

        eigen = nx.eigenvector_centrality(self.G)
        print('eigenvector centrality\n')
        eigen_l = sorted(eigen.items(), key=operator.itemgetter(1), reverse=True)
        for n in eigen_l:
            print(str(n))
        print()

        betweenness = nx.betweenness_centrality(self.G)
        print('betweenness centrality\n')
        betweenness_l = sorted(betweenness.items(), key=operator.itemgetter(1), reverse=True)
        for n in betweenness_l:
            print(str(n))
        print()

    def save_model(self, embedding_filename, embedding_model_file):
        # Save embeddings for later use
        self.model.wv.save_word2vec_format(embedding_filename, binary=True)

        # Save model for later use
        self.model.save(embedding_model_file)

    def load_model(self, embedding_filename):
        self.model = KeyedVectors.load_word2vec_format(embedding_filename, binary=True, unicode_errors='ignore')

    #to do: node2vec model loading...

    def most_similars(self, word):
        # Look for most similar nodes
        return self.model.most_similar(word)  # Output node names are always strings

    def compute_similarity(self, first_node, second_node):
        edges_embs = HadamardEmbedder(keyed_vectors=self.model)

        # Look for embeddings on the fly - here we pass normal tuples
        edges_embs[(first_node, second_node)]
        ''' OUTPUT
        array([ 5.75068220e-03, -1.10937878e-02,  3.76693785e-01,  2.69105062e-02,
               ... ... ....
               ..................................................................],
              dtype=float32)
        '''

        # Get all edges in a separate KeyedVectors instance - use with caution could be huge for big networks
        edges_kv = edges_embs.as_keyed_vectors()

        try:
            # Look for most similar edges - this time tuples must be sorted and as str
            results = edges_kv.most_similar(str((first_node, second_node)))

            # Save embeddings for later use
            # edges_kv.save_word2vec_format(EDGES_EMBEDDING_FILENAME)
        except:
            return []

        return results


import numpy as np
from .utils import *


class Word2Vec:
    """
    Python implementation of Word2Vec.
    # Arguments
        method : `str`
            choose method for word2vec (options: 'cbow', 'skipgram')
            [default: 'cbow']
        window_size: `integer`
            size of window [default: 1]
        n_hidden: `integer`
            size of hidden layer [default: 2]
        n_epochs: `integer`
            number of epochs [default: 1]
        learning_rate: `float` [default: 0.1]
        corpus: `str`
            corpus text
    """

    def __init__(self, method='cbow', window_size=1, n_hidden=2, n_epochs=1, corpus='', learning_rate=0.1):
        self.window = window_size
        self.N = n_hidden
        self.n_epochs = n_epochs
        self.corpus = [corpus]
        self.eta = learning_rate
        if method == 'cbow':
            self.method = self.cbow
        elif method == 'skipgram':
            self.method = self.skipgram
        else:
            raise ValueError("Method not recognized. Aborting.")

    def cbow(self, context, label, W1, W2, loss):
        """
        Implementation of Continuous-Bag-of-Words Word2Vec model
        :param context: all the context words (these represent the inputs)
        :param label: the center word (this represents the label)
        :param W1: weights from the input to the hidden layer
        :param W2: weights from the hidden to the output layer
        :param loss: float that represents the current value of the loss function
        :return: updated weights and loss
        """
        # context is 'x' from tokenizer, it is a c x V matrix
        # label is 'y' from tokenizer, it is a 1 x V matrix

        x = np.matrix(np.mean(context, axis=0))

        # x is a 1 x V matrix
        # W1 is a VxN matrix
        # h is a N x 1 matrix
        h = np.matmul(W1.T, x.T)

        # u is a V x 1 matrix
        u = np.matmul(W2.T, h)

        # W2 is an N x V matrix
        # y_pred is a V x 1 matrix
        y_pred = softmax(u)
        # e is a V x 1 matrix
        e = -label.T + y_pred
        # h is N x 1 and e is V x 1 so dW2 is N x V
        dW2 = np.outer(h, e)
        # x.T is a V x 1 matrix, W2e is a Nx1 so dW1 this is V x N
        dW1 = np.outer(x.T, np.matmul(W2, e))

        new_W1 = W1 - self.eta * dW1
        new_W2 = W2 - self.eta * dW2

        # label is a 1xV matrix so label.T is a Vx1 matrix
        loss += -float(u[label.T == 1]) + np.log(np.sum(np.exp(u)))

        return new_W1, new_W2, loss

    def skipgram(self, context, x, W1, W2, loss):
        """
        Implementation of Skip-Gram Word2Vec model
        :param context: all the context words (these represent the labels)
        :param x: the center word (this represents the input)
        :param W1: weights from the input to the hidden layer
        :param W2: weights from the hidden to the output layer
        :param loss: float that represents the current value of the loss function
        :return: updated weights and loss
        """
        # context is "x" from tokenizer, it is a c x V matrix
        # "x" is "y" from tokenizer, it is a 1 x V matrix
        # W1 has dimension V x N (N= number of features, V = vocab size)
        # x has dimension V x 1
        h = np.matmul(W1.T, x.T)
        # h has dimension N x 1
        # W2 has dimension N x V
        # u has dimension V x 1
        u = np.dot(W2.T, h)
        # y_pred has dimension V x 1
        y_pred = softmax(u)

        # context is a c by V matrix
        # e is a V x c matrix
        e = np.outer(y_pred, np.array([1] * context.shape[0])) - context.T

        # np.sum(e, axis=1) is a V x 1 vectors
        # h is an N x 1 Vector
        # dW2 is a N x V matrix
        dW2 = np.outer(h, np.sum(e, axis=1))
        # x is a V x 1 matrix
        # np.dot(W2, np.sum(e,axis=1)) is a product (N x V) (Vx 1) is Nx1
        # dW1 is an V x N matrix
        dW1 = np.outer(x, np.dot(W2, np.sum(e, axis=1)))

        new_W1 = W1 - self.eta * dW1
        new_W2 = W2 - self.eta * dW2

        loss += - np.sum([u[label.T == 1] for label in context]) + len(context) * np.log(np.sum(np.exp(u)))

        return new_W1, new_W2, loss

    def predict(self, x, W1, W2):
        """Predict output from input data and weights
        :param x: input data
        :param W1: weights from input to hidden layer
        :param W2: weights from hidden layer to output layer
        :return: output of neural network
        """
        h = np.mean([np.matmul(W1.T, xx) for xx in x], axis=0)
        u = np.dot(W2.T, h)
        return softmax(u)

    def run(self):
        """
        Main method of the Word2Vec class.
        :return: the final values of the weights W1, W2 and a history of the value of the loss function vs. epoch
        """
        if len(self.corpus) == 0:
            raise ValueError('You need to specify a corpus of text.')

        corpus_tokenized, V = tokenize(self.corpus)
        W1, W2 = initialize(V, self.N)

        loss_vs_epoch = []
        for e in range(self.n_epochs):
            loss = 0.
            for context, center in corpus2io(corpus_tokenized, V, self.window):
                W1, W2, loss = self.method(context, center, W1, W2, loss)
            loss_vs_epoch.append(loss)

        return W1, W2, loss_vs_epoch


from yonlu.model import *
from yonlu.relation_classification import *
from yonlu.data_utils import *
from yonlu.word_embeddings import *
from yonlu.kobert import *
from yonlu.ner import *
from yonlu.machine_translation import *
from yonlu.bert import *



import torch
import numpy as np
from torch import nn, optim
from transformers import BertModel, BertTokenizer, AdamW, get_linear_schedule_with_warmup
from collections import defaultdict
import torch.nn.functional as F
import time
import datetime
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import os

class PYBERTTrainer:
    def __init__(self):
        print('training model with BERT')

        # We'll store a number of quantities such as training and validation loss,
        # validation accuracy, and timings.
        self.training_stats = []

        self.df_stats = []

    def format_time(self, elapsed):
        '''
        Takes a time in seconds and returns a string hh:mm:ss
        '''
        # Round to the nearest second.
        elapsed_rounded = int(round((elapsed)))

        # Format as hh:mm:ss
        return str(datetime.timedelta(seconds=elapsed_rounded))

    def train_epoch(self, model, data_loader, loss_fn, optimizer, device, scheduler, algorithm='transformers'):
        model = model.train()
        correct_predictions = 0.0
        # Reset the total loss for this epoch.
        total_train_loss = 0.0

        for d in data_loader:
            input_ids = d["input_ids"].to(device)
            attention_mask = d["attention_mask"].to(device)
            token_type_ids = d["token_type_ids"].to(device)
            targets = d["targets"].to(device)

            if algorithm == 'transformers':
                # Always clear any previously calculated gradients before performing a
                # backward pass. PyTorch doesn't do this automatically because
                # accumulating the gradients is "convenient while training RNNs".
                # (source: https://stackoverflow.com/questions/48001598/why-do-we-need-to-call-zero-grad-in-pytorch)
                model.zero_grad()

                loss, logits = model(input_ids=input_ids,
                                     token_type_ids=token_type_ids,
                                     attention_mask=attention_mask,
                                     labels=targets, return_dict=False)

                pred = torch.argmax(F.softmax(logits), dim=1)
                correct = pred.eq(targets)

                label_ids = targets.to('cpu').numpy()

                correct_predictions += correct.sum().item()

                # Accumulate the training loss over all of the batches so that we can
                # calculate the average loss at the end. `loss` is a Tensor containing a
                # single value; the `.item()` function just returns the Python value
                # from the tensor.
                total_train_loss += loss.item()

                # Perform a backward pass to calculate the gradients.
                loss.backward()

                # Clip the norm of the gradients to 1.0.
                # This is to help prevent the "exploding gradients" problem.
                torch.nn.utils.clip_grad_norm_(model.parameters(), 1.0)

                # Update parameters and take a step using the computed gradient.
                # The optimizer dictates the "update rule"--how the parameters are
                # modified based on their gradients, the learning rate, etc.
                optimizer.step()

                # Update the learning rate.
                scheduler.step()

            else:
                #print('starting the process based on BertModel API, which is a harder way')
                #using BertModel API directly
                if model.name() == 'PYBERTClassifier':
                    outputs = model(
                        input_ids=input_ids,
                        attention_mask=attention_mask
                    )
                elif model.name() == 'PYBERTClassifierGenAtten':
                    outputs = model(input_ids, targets, token_type_ids,attention_mask)

                _, preds = torch.max(outputs, dim=1)
                loss = loss_fn(outputs, targets)
                correct_predictions += torch.sum(preds == targets)

                total_train_loss += loss.item()
                loss.backward()
                nn.utils.clip_grad_norm_(model.parameters(), max_norm=1.0)
                optimizer.step()
                scheduler.step()
                optimizer.zero_grad()

        avg_train_avg = float(correct_predictions) / float(len(data_loader.dataset))*100
        # Calculate the average loss over all of the batches.
        avg_train_loss = total_train_loss / len(data_loader)

        return avg_train_avg, avg_train_loss

    def flat_accuracy(self, preds, labels):
        '''
         # Function to calculate the accuracy of our predictions vs labels
        :param preds:
        :param labels:
        :return:
        '''
        pred_flat = np.argmax(preds, axis=1).flatten()
        labels_flat = labels.flatten()
        return np.sum(pred_flat == labels_flat) / len(labels_flat)

    def eval_model(self, model, data_loader, loss_fn, device, algorithm='transformers'):
        model = model.eval()

        correct_predictions = 0
        total_eval_accuracy = 0.0
        total_eval_loss = 0.0

        with torch.no_grad():
            for d in data_loader:
                input_ids = d["input_ids"].to(device)
                attention_mask = d["attention_mask"].to(device)
                token_type_ids = d["token_type_ids"].to(device)
                targets = d["targets"].to(device)

                if algorithm == 'transformers':
                    (loss, logits) = model(input_ids,
                                           token_type_ids=None,
                                           attention_mask=attention_mask,
                                           labels=targets, return_dict=False)

                    # Accumulate the validation loss.
                    total_eval_loss += loss.item()
                    # Move logits and labels to CPU
                    logits = logits.detach().cpu().numpy()
                    label_ids = targets.to('cpu').numpy()

                    # Calculate the accuracy for this batch of test sentences, and
                    # accumulate it over all batches.
                    total_eval_accuracy += self.flat_accuracy(logits, label_ids)

                else:
                    if model.name() == 'PYBERTClassifier':
                        outputs = model(
                            input_ids=input_ids,
                            attention_mask=attention_mask
                        )
                    elif model.name() == 'PYBERTClassifierGenAtten':
                        outputs = model(
                            input_ids,
                            targets,
                            token_type_ids,
                            attention_mask)

                    _, preds = torch.max(outputs, dim=1)
                    loss = loss_fn(outputs, targets)

                    total_eval_accuracy += torch.sum(preds == targets)
                    # Accumulate the validation loss.
                    total_eval_loss += loss.item()

        # Report the final accuracy for this validation run.
        avg_val_accuracy = float(total_eval_accuracy) / float(len(data_loader))*10

        # Calculate the average loss over all of the batches.
        avg_val_loss = total_eval_loss / len(data_loader)

        return avg_val_accuracy, avg_val_loss

    def train(self, model, device, train_data_loader, val_data_loader, df_val,
              df_train, tokenizer, num_epochs=10, algorithm='transformers',
              torch_model_name='best_model_states.bin'):

        # Measure the total training time for the whole run.
        total_t0 = time.time()

        # Note: AdamW is a class from the huggingface library (as opposed to pytorch)
        # I believe the 'W' stands for 'Weight Decay fix"
        optimizer = AdamW(model.parameters(),
                          lr=2e-5,  # args.learning_rate - default is 5e-5, our notebook had 2e-5
                          eps=1e-8  # args.adam_epsilon  - default is 1e-8.
                          )

        total_steps = len(train_data_loader) * num_epochs

        # Create the learning rate scheduler.
        scheduler = get_linear_schedule_with_warmup(
            optimizer,
            num_warmup_steps=0,  # Default value in run_glue.py
            num_training_steps=total_steps
        )
        loss_fn = nn.CrossEntropyLoss().to(device)

        best_accuracy = 0

        # ========================================
        #               Training
        # ========================================
        # Perform one full pass over the training set.
        for epoch in range(num_epochs):
            print("")
            print('======== Epoch {:} / {:} ========'.format(epoch + 1, num_epochs))
            print('Training...')

            # Measure how long the training epoch takes.
            t0 = time.time()

            train_acc, train_loss = self.train_epoch(
                model,
                train_data_loader,
                loss_fn,
                optimizer,
                device,
                scheduler,
                algorithm=algorithm
            )

            # Measure how long this epoch took.
            training_time = self.format_time(time.time() - t0)

            print("")
            print("  Average training loss: {0:.2f}".format(train_loss))
            print("  Average accuracy: {0:.2f}".format(train_acc))

            print("  Training epcoh took: {:}".format(training_time))

            # ========================================
            #               Validation
            # ========================================
            # After the completion of each training epoch, measure our performance on
            # our validation set.

            print("")
            print("Running Validation...")
            t0 = time.time()

            val_acc, val_loss = self.eval_model(
                model,
                val_data_loader,
                loss_fn,
                device,
                algorithm=algorithm
            )

            # Measure how long the validation run took.
            validation_time = self.format_time(time.time() - t0)

            print("  Validation Loss: {0:.2f}".format(val_loss))
            print("  Validation took: {:}".format(validation_time))

            # Record all statistics from this epoch.
            self.training_stats.append(
                {
                    'epoch': epoch + 1,
                    'Training Loss': train_loss,
                    'Valid. Loss': val_loss,
                    'Valid. Accur.': val_acc,
                    'Training Time': training_time,
                    'Validation Time': validation_time
                }
            )

            print("")
            print("Training complete!")

            print("Total training took {:} (h:mm:ss)".format(self.format_time(time.time() - total_t0)))

            if val_acc > best_accuracy:
                #torch.save(model.state_dict(), saved_training_model)
                self._save(model,tokenizer, algorithm, torch_model_name)
                best_accuracy = val_acc


    def _save(self, model, tokenizer, algorithm, torch_model_name='best_model_states.bin'):
        # Saving best-practices: if you use defaults names for the model, you can reload it using from_pretrained()

        output_dir = './model_save/'

        # Create output directory if needed
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        print("Saving model to %s" % output_dir)


        # Save a trained model, configuration and tokenizer using `save_pretrained()`.
        # They can then be reloaded using `from_pretrained()`
        #model_to_save = model.module if hasattr(model, 'module') else model  # Take care of distributed/parallel training
        #model_to_save.save_pretrained(output_dir)
        torch.save(model, os.path.join(output_dir, torch_model_name))


    def summanry_training_stats(self):
        # Display floats with two decimal places.
        pd.set_option('precision', 2)

        # Create a DataFrame from our training statistics.
        df_stats = pd.DataFrame(data=self.training_stats)

        # Use the 'epoch' as the row index.
        self.df_stats = df_stats.set_index('epoch')

        # A hack to force the column headers to wrap.
        # df = df.style.set_table_styles([dict(selector="th",props=[('max-width', '70px')])])

        # Display the table.
        print(self.df_stats)

    def visualize_performance(self):
        # Use plot styling from seaborn.
        sns.set(style='darkgrid')

        # Increase the plot size and font size.
        sns.set(font_scale=1.5)
        plt.rcParams["figure.figsize"] = (12, 6)

        # Plot the learning curve.
        plt.plot(self.df_stats['Training Loss'], 'b-o', label="Training")
        plt.plot(self.df_stats['Valid. Loss'], 'g-o', label="Validation")

        # Label the plot.
        plt.title("Training & Validation Loss")
        plt.xlabel("Epoch")
        plt.ylabel("Loss")
        plt.legend()
        plt.xticks([1, 2, 3, 4])

        plt.show()


import os
import openai
import glob
from yonlu.bert.gpt3 import GPT
from yonlu.bert.gpt3 import Example


openai.api_key = "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"

gpt = GPT(engine="davinci",
          temperature=0.5,
          output_prefix="Output: \n\n",
          max_tokens=100)

# add some code examples
for file in glob.glob("examples/*"):
    title = file.replace("_", " ")
    with open(f"{file}", "r") as f:
        code = f.read()
    gpt.add_example(Example(title, code))

# add some calculation examples
gpt.add_example(Example("add 3+5", "8"))
gpt.add_example(Example("add 8+5", "13"))
gpt.add_example(Example("add 50+25", "75"))

# Inferences
prompt = "sort list in python"
output = gpt.get_top_reply(prompt)
print(prompt, ":", output)
print("----------------------------------------")

prompt = "Code weather api in python"
output = gpt.get_top_reply(prompt)
print(prompt, ":", output)
print("----------------------------------------")

prompt = "What is 876+89"
output = gpt.get_top_reply(prompt)
print(prompt, ":", output)
print("----------------------------------------")



import os

from keras.models import Sequential
from keras.layers import Embedding, Flatten, Dense
from keras.preprocessing.sequence import pad_sequences
from keras.models import load_model

from tqdm import tqdm
import numpy as np
from keras.preprocessing.text import Tokenizer

class word2vecFilteringFunction:
    def __init__(self):

        self.maxlen = 100 # Make all sequences 100 words long
        self.embedding_dim = 100  # We now use larger embeddings
        self.max_words = 10000 # We will only consider the 10K most used words in this dataset

        self.tokenizer = Tokenizer(num_words=self.max_words)

        self.labels = []

        # deep learning model
        self.model = Sequential()

    def fetch(self):

        imdb_dir = 'D:\\My Documents\\teaching\\nlp_deeplearning\\aclImdb_v1\\aclImdb'  # Data directory
        train_dir = os.path.join(imdb_dir, 'train')  # Get the path of the train set

        # Setup empty lists to fill
        labels = []
        texts = []

        # First go through the negatives, then through the positives
        for label_type in ['neg', 'pos']:
            # Get the sub path
            dir_name = os.path.join(train_dir, label_type)
            print('loading ', label_type)
            # Loop over all files in path
            for fname in tqdm(os.listdir(dir_name)):
                # Only consider text files
                if fname[-4:] == '.txt':
                    # Read the text file and put it in the list
                    f = open(os.path.join(dir_name, fname))
                    texts.append(f.read())
                    f.close()
                    # Attach the corresponding label
                    if label_type == 'neg':
                        labels.append(0)
                    else:
                        labels.append(1)

        self.labels = labels;

        # print mean value
        print('label mean value' + str(np.mean(labels)))

        #print sample text and label
        print('Label',labels[24002])
        print(texts[24002])

        return texts

    def tokenize(self, texts):

        self.tokenizer.fit_on_texts(texts) # Generate tokens by counting frequency
        sequences = self.tokenizer.texts_to_sequences(texts) # Turn text into sequence of numbers
        word_index = self.tokenizer.word_index
        print('Token for "the"',word_index['the'])
        print('Token for "Movie"',word_index['movie'])
        print('Token for "generator"',word_index['generator'])

        # Display the first 10 words of the sequence tokenized
        print (str(sequences[24002][:10]))

        return sequences


    def createEmbeddings(self):
        glove_dir = '../embeddings' # This is the folder with the dataset

        print('Loading word vectors')
        embeddings_index = {} # We create a dictionary of word -> embedding
        f = open(os.path.join(glove_dir, 'glove.6B.100d.txt')) # Open file

        # In the dataset, each line represents a new word embedding
        # The line starts with the word and the embedding values follow
        for line in tqdm(f):
            values = line.split()
            word = values[0] # The first value is the word, the rest are the values of the embedding
            embedding = np.asarray(values[1:], dtype='float32') # Load embedding
            embeddings_index[word] = embedding # Add embedding to our embedding dictionary
        f.close()

        print('Found %s word vectors.' % len(embeddings_index))

        # Create a matrix of all embeddings
        all_embs = np.stack(embeddings_index.values())
        emb_mean = all_embs.mean() # Calculate mean
        emb_std = all_embs.std() # Calculate standard deviation

        #filter input data by word embedding

        word_index = self.tokenizer.word_index
        nb_words = min(self.max_words, len(word_index)) # How many words are there actually

        # Create a random matrix with the same mean and std as the embeddings
        embedding_matrix = np.random.normal(emb_mean, emb_std, (nb_words, self.embedding_dim))

        # The vectors need to be in the same position as their index.
        # Meaning a word with token 1 needs to be in the second row (rows start with zero) and so on

        # Loop over all words in the word index
        for word, i in word_index.items():
            # If we are above the amount of words we want to use we do nothing
            if i >= self.max_words:
                continue
            # Get the embedding vector for the word
            embedding_vector = embeddings_index.get(word)
            # If there is an embedding vector, put it in the embedding matrix
            if embedding_vector is not None:
                embedding_matrix[i] = embedding_vector

        return embedding_matrix


    def buildModel(self, filtering=True):

        if (filtering):
            embedding_matrix = self.createEmbeddings();
            self.model.add(Embedding(self.max_words, self.embedding_dim, input_length=self.maxlen, weights = [embedding_matrix], trainable = False))
        else:
            self.model.add(Embedding(self.max_words, self.embedding_dim, input_length=self.maxlen))


        self.model.add(Flatten())
        self.model.add(Dense(32, activation='relu'))
        self.model.add(Dense(1, activation='sigmoid'))
        self.model.summary()

    def train(self, sequences):
        #training and validation data

        data = pad_sequences(sequences, maxlen=self.maxlen)
        print(data.shape) # We have 25K, 100 word sequences now

        labels = np.asarray(self.labels)

        # Shuffle data
        indices = np.arange(data.shape[0])
        np.random.shuffle(indices)
        data = data[indices]
        labels = labels[indices]

        training_samples = 20000  # We will be training on 10K samples
        validation_samples = 5000  # We will be validating on 10000 samples

        # Split data
        x_train = data[:training_samples]
        y_train = labels[:training_samples]
        x_val = data[training_samples: training_samples + validation_samples]
        y_val = labels[training_samples: training_samples + validation_samples]

        #
        self.model.compile(optimizer='adam',
                      loss='binary_crossentropy',
                      metrics=['acc'])

        history = self.model.fit(x_train, y_train,
                            epochs=10,
                            batch_size=32,
                            validation_data=(x_val, y_val))

    def saveModel(self, model_file):
        self.model.save(model_file, overwrite=True, include_optimizer=True)


    def loadModel(self, model_file):
        self.model = load_model(model_file)

    def test(self):
        # Demo on a positive text
        my_text = 'I love dogs. Dogs are the best. They are lovely, cuddly animals that only want the best for humans.'

        seq = self.tokenizer.texts_to_sequences([my_text])
        print('raw seq:', seq)
        seq = pad_sequences(seq, maxlen=self.maxlen)
        print('padded seq:', seq)
        prediction = self.model.predict(seq)
        print('positivity:', prediction)


if __name__ == '__main__':

    word2vec_filter = word2vecFilteringFunction()

    model_file = 'sentiment_model.model'
    # fetch data
    texts = word2vec_filter.fetch()

    # tokenize data
    sequences = word2vec_filter.tokenize(texts)

    mode = 'test' # train or test

    if (mode == 'train'):
        filtering_mode = True
        #create deel learning model for sentiment analysis
        word2vec_filter.buildModel(filtering_mode)

        #train the classifier
        word2vec_filter.train(sequences)

        word2vec_filter.saveModel(model_file)

    elif (mode == 'test'):
        #load
        word2vec_filter.loadModel(model_file)

        word2vec_filter.test()


import matplotlib.pyplot as plt
import seaborn as sns

from torch.utils.data import TensorDataset, random_split
from torch.utils.data import DataLoader, RandomSampler, SequentialSampler
import torch
from sklearn.metrics import matthews_corrcoef
import numpy as np
import torch.nn.functional as F

from yonlu.bert.bert_classification_model import PYBERTClassifier
from yonlu.bert.bert_dataset import PYBERTDataset


class bert_predictor:

    def __init(self):
        print('BERT Predictor')
        self.prediction_dataloader = None
        self.model = None

    def load_data(self, df, tokenizer, max_len):
        # Create sentence and label lists
        sentences = df.content.values
        labels = df.sentiment.values

        # Tokenize all of the sentences and map the tokens to thier word IDs.
        input_ids = []
        attention_masks = []

        # For every sentence...
        for sent in sentences:
            # `encode_plus` will:
            #   (1) Tokenize the sentence.
            #   (2) Prepend the `[CLS]` token to the start.
            #   (3) Append the `[SEP]` token to the end.
            #   (4) Map tokens to their IDs.
            #   (5) Pad or truncate the sentence to `max_length`
            #   (6) Create attention masks for [PAD] tokens.
            encoded_dict = tokenizer.encode_plus(
                sent,  # Sentence to encode.
                add_special_tokens=True,  # Add '[CLS]' and '[SEP]'
                max_length=max_len,  # Pad & truncate all sentences.
                pad_to_max_length=True,
                return_attention_mask=True,  # Construct attn. masks.
                return_tensors='pt',  # Return pytorch tensors.
            )

            # Add the encoded sentence to the list.
            input_ids.append(encoded_dict['input_ids'])

            # And its attention mask (simply differentiates padding from non-padding).
            attention_masks.append(encoded_dict['attention_mask'])

        # Convert the lists into tensors.
        input_ids = torch.cat(input_ids, dim=0)
        attention_masks = torch.cat(attention_masks, dim=0)
        labels = torch.tensor(labels)

        # Set the batch size.
        batch_size = 32

        # Create the DataLoader.
        #prediction_data = TensorDataset(input_ids, attention_masks, labels)
        #prediction_sampler = SequentialSampler(prediction_data)

        ds = PYBERTDataset(
            contents=df.content.to_numpy(),
            targets=df.sentiment.to_numpy(),
            tokenizer=tokenizer,
            max_len=max_len)

        prediction_sampler = SequentialSampler(ds)

        self.prediction_dataloader = DataLoader(ds,
                                                sampler=prediction_sampler,
                                                batch_size=batch_size)

        # Prediction on test set
        print('Predicting labels for {:,} test sentences...'.format(len(input_ids)))

        print('Positive samples: %d of %d (%.2f%%)' % (
        df.sentiment.sum(), len(df.sentiment), (df.sentiment.sum() / len(df.sentiment) * 100.0)))

    def load_model(self, saved_training_model):
        self.model = torch.load(saved_training_model)

        # Put model in evaluation mode
        self.model.eval()

    def predict(self, device, algorithm='transformers'):
        contents = []
        predictions = []
        prediction_probs = []
        true_labels = []

        # Predict
        for batch in self.prediction_dataloader:

            # batch = tuple(t.to(device) for t in batch)
            # Unpack the inputs from our dataloader
            texts = batch["document_text"]
            input_ids = batch["input_ids"].to(device)
            attention_mask = batch["attention_mask"].to(device)
            targets = batch["targets"].to(device)
            token_type_ids = batch["token_type_ids"].to(device)
            #print(str(len(input_ids)) + ' == ' + str(len(targets)))

            # Telling the model not to compute or store gradients, saving memory and
            # speeding up prediction
            with torch.no_grad():
                # Forward pass, calculate logit predictions
                if algorithm == 'transformers':
                    outputs = self.model(input_ids,
                                         token_type_ids=None,
                                         attention_mask=attention_mask,
                                         return_dict=False)
                else:
                    if self.model.name() == 'PYBERTClassifier':
                        outputs = self.model(
                            input_ids=input_ids,
                            attention_mask=attention_mask
                        )
                    elif self.model.name() == 'PYBERTClassifierGenAtten':
                        outputs = self.model(input_ids, targets, token_type_ids, attention_mask)
                    #outputs = self.model(
                    #    input_ids=input_ids,
                    #    attention_mask=attention_mask
                    #)

            if algorithm == 'transformers':
                logits = outputs[0]
                _, preds = torch.max(logits, dim=1)
                # Move logits and labels to CPU
                preds = preds.detach().cpu().numpy()
                # The predictions for this batch are a 2-column ndarray (one column for "0"
                # and one column for "1"). Pick the label with the highest value and turn this
                # in to a list of 0s and 1s.
                #probs = F.softmax(logits, dim=1)
                probs = logits
            else:
                _, preds = torch.max(outputs, dim=1)
                preds = preds.detach().cpu().numpy()
                probs = outputs

            targets = targets.to('cpu').numpy()

            # Store predictions and true labels along with text and prediction probability
            # need to know the difference between python list's extend and append
            contents.extend(texts)
            predictions.extend(preds)
            prediction_probs.extend(probs)
            true_labels.extend(targets)

        print(' DONE.')

        return contents, predictions, prediction_probs, true_labels

    def predict_each(self, device, text, tokenizer, MAX_LEN, algorithm='transformers'):

        encoded_text = tokenizer.encode_plus(
            text,
            max_length=MAX_LEN,
            add_special_tokens=True,
            return_token_type_ids=True,
            pad_to_max_length=True,
            return_attention_mask=True,
            return_tensors='pt',
        )

        input_ids = encoded_text['input_ids'].to(device)
        attention_mask = encoded_text['attention_mask'].to(device)
        token_type_ids = encoded_text["token_type_ids"].to(device)

        # Predict
        preds = 0
        probs = None

        # Forward pass, calculate logit predictions
        if algorithm == 'transformers':
            outputs = self.model(input_ids,
                                 token_type_ids=None,
                                 attention_mask=attention_mask)
        else:
            if self.model.name() == 'PYBERTClassifier':
                outputs = self.model(
                    input_ids=input_ids,
                    attention_mask=attention_mask
                )
            elif self.model.name() == 'PYBERTClassifierGenAtten':

                outputs = self.model(input_ids, None, token_type_ids, attention_mask)

            #outputs = self.model(
            #    input_ids=input_ids,
            #    attention_mask=attention_mask
            #)

        if algorithm == 'transformers':
            logits = outputs[0]
            _, preds = torch.max(logits, dim=1)
            # Move logits and labels to CPU
            preds = preds.detach().cpu().numpy()
            # The predictions for this batch are a 2-column ndarray (one column for "0"
            # and one column for "1"). Pick the label with the highest value and turn this
            # in to a list of 0s and 1s.
            # probs = F.softmax(logits, dim=1)
            probs = logits
        else:
            _, preds = torch.max(outputs, dim=1)
            preds = preds.detach().cpu().numpy()
            probs = outputs

        print(str(preds) + " :: " + str(probs))
        return preds[0]

    def evaluate(self, labels, predictions):
        matthews_set = []

        # Evaluate each test batch using Matthew's correlation coefficient
        print('Calculating Matthews Corr. Coef. for each batch...')
        # For each input batch...
        for i in range(len(labels)):
            # The predictions for this batch are a 2-column ndarray (one column for "0"
            # and one column for "1"). Pick the label with the highest value and turn this
            # in to a list of 0s and 1s.
            #pred_labels_i = np.argmax(predictions[i], axis=1).flatten()
            pred_labels_i = predictions[i]
            # Calculate and store the coef for this batch.
            matthews = matthews_corrcoef(labels[i], pred_labels_i)
            matthews_set.append(matthews)

        # Create a barplot showing the MCC score for each batch of test samples.
        ax = sns.barplot(x=list(range(len(matthews_set))), y=matthews_set, ci=None)

        plt.title('MCC Score per Batch')
        plt.ylabel('MCC Score (-1 to +1)')
        plt.xlabel('Batch #')

        plt.show()

        # Combine the results across all batches.
        flat_predictions = np.concatenate(predictions, axis=0)
        # For each sample, pick the label (0 or 1) with the higher score.
        flat_predictions = np.argmax(flat_predictions, axis=1).flatten()

        # Combine the correct labels for each batch into a single list.
        flat_true_labels = np.concatenate(labels, axis=0)

        # Calculate the MCC
        mcc = matthews_corrcoef(flat_true_labels, flat_predictions)

        print('Total MCC: %.3f' % mcc)




import warnings

warnings.filterwarnings("ignore")
import sys

import torch
from yonlu.model.transformer import Transformer
from transformers import BertTokenizer
from yonlu.data_utils.translation_dataset import TranslationDataset
from torch.utils.data import DataLoader, random_split
from tqdm import tqdm
import math
import os
import time


class TranslationTrainer():
  def __init__(self,
               dataset,
               tokenizer,
               model,
               max_len,
               device,
               model_name,
               checkpoint_path,
               batch_size,
               ):
    self.dataset = dataset
    self.tokenizer = tokenizer
    self.model = model
    self.max_len = max_len
    self.model_name = model_name
    self.checkpoint_path = checkpoint_path
    self.device = device
    self.ntoken = tokenizer.vocab_size
    self.batch_size = batch_size

  def my_collate_fn(self, samples):
    input_str =[]
    target_str =[]
    input_str.append([sample['input_str'] for sample in samples])
    input = [sample['input'] for sample in samples]
    input_mask = [sample['input_mask'] for sample in samples]
    target = [sample['target'] for sample in samples]
    target_mask = [sample['target_mask'] for sample in samples]
    token_num = [sample['token_num'] for sample in samples]
    target_str.append([sample['target_str'] for sample in samples])

    return {
        "input_str":input_str,
        "input":torch.stack(input).contiguous(),                                              # input
        "input_mask": torch.stack(input_mask).contiguous(),       # input_mask
        "target": torch.stack(target).contiguous(),                                           # target,
        "target_mask": torch.stack(target_mask).contiguous(),   # target_mask
        "token_num": torch.stack(token_num).contiguous(),   # token_num
        "target_str": target_str
    }

  def build_dataloaders(self, train_test_split=0.1, train_shuffle=True, eval_shuffle=True):
    dataset_len = len(self.dataset)
    eval_len = int(dataset_len * train_test_split)
    train_len = dataset_len - eval_len
    train_dataset, eval_dataset = random_split(self.dataset, (train_len, eval_len))
    train_loader = DataLoader(train_dataset, batch_size=self.batch_size, shuffle=train_shuffle, collate_fn=self.my_collate_fn)
    eval_loader = DataLoader(eval_dataset, batch_size=self.batch_size, shuffle=eval_shuffle , collate_fn=self.my_collate_fn)

    return train_loader, eval_loader

  def train(self, epochs, train_dataset, eval_dataset, optimizer, scheduler):
    self.model.train()
    total_loss = 0.
    global_steps = 0
    start_time = time.time()
    losses = {}
    best_val_loss = float("inf")
    best_model = None
    start_epoch = 0
    start_step = 0
    train_dataset_length = len(train_dataset)

    self.model.to(self.device)
    if os.path.isfile(f'{self.checkpoint_path}/{self.model_name}.pth'):
      checkpoint = torch.load(f'{self.checkpoint_path}/{self.model_name}.pth', map_location=self.device)
      start_epoch = checkpoint['epoch']
      losses = checkpoint['losses']
      global_steps = checkpoint['train_step']
      start_step = global_steps if start_epoch == 0 else (global_steps % train_dataset_length) + 1

      self.model.load_state_dict(checkpoint['model_state_dict'])
      optimizer.load_state_dict(checkpoint['optimizer_state_dict'])

    for epoch in range(start_epoch, epochs):
      epoch_start_time = time.time()

      pb = tqdm(enumerate(train_dataset),
                desc=f'Epoch-{epoch} Iterator',
                total=train_dataset_length,
                bar_format='{l_bar}{bar:10}{r_bar}'
                )
      pb.update(start_step)
      for i,data in pb:
        if i < start_step:
          continue
        """
        doc={
        "input":input,                                              # input
        "input_mask": (input != pad_token_idx).unsqueeze(-2),       # input_mask
        "target": target,                                           # target,
        "target_mask": self.make_std_mask(target, pad_token_idx),   # target_mask
        "token_num": (target[...,1:] != pad_token_idx).data.sum()   # token_num
      }
        """
        input = data['input'].to(self.device)
        target = data['target'].to(self.device)
        input_mask = data['input_mask'].to(self.device)
        target_mask = data['target_mask'].to(self.device)

        optimizer.zero_grad()
        generator_logit, loss = self.model.forward(input, target, input_mask, target_mask, labels=target)

        loss.backward()
        torch.nn.utils.clip_grad_norm_(self.model.parameters(), 0.5)
        optimizer.step()

        losses[global_steps] = loss.item()
        total_loss += loss.item()
        log_interval = 1
        save_interval = 3

        global_steps += 1

        if i % log_interval == 0 and i > 0:
          cur_loss = total_loss / log_interval
          elapsed = time.time() - start_time
          # print('| epoch {:3d} | {:5d}/{:5d} batches | '
          #       'lr {:02.2f} | ms/batch {:5.2f} | '
          #       'loss {:5.2f} | ppl {:8.2f}'.format(
          #         epoch, i, len(train_dataset), scheduler.get_lr()[0],
          #         elapsed * 1000 / log_interval,
          #         cur_loss, math.exp(cur_loss)))
          pb.set_postfix_str('| epoch {:3d} | {:5d}/{:5d} batches | '
                             'lr {:02.2f} | ms/batch {:5.2f} | '
                             'loss {:5.2f} | ppl {:8.2f}'.format(
            epoch, i, len(train_dataset), scheduler.get_lr()[0],
            elapsed * 1000 / log_interval,
            cur_loss, math.exp(cur_loss)))
          total_loss = 0
          start_time = time.time()
          if i % save_interval == 0:
            self.save(epoch, self.model, optimizer, losses, global_steps)
      val_loss = self.evaluate(eval_dataset)
      self.model.train()
      print('-' * 89)
      print('| end of epoch {:3d} | time: {:5.2f}s | valid loss {:5.2f} | '
            'valid ppl {:8.2f}'.format(epoch, (time.time() - epoch_start_time),
                                       val_loss, math.exp(val_loss)))
      print('-' * 89)
      if val_loss < best_val_loss:
        best_val_loss = val_loss
        #best_model = self.model
        self.save(epoch, self.model, optimizer, losses, global_steps)

      start_step = 0
      scheduler.step()

  def evaluate(self, dataset):
    self.model.eval()  # 평가 모드를 시작합니다.
    total_loss = 0.

    self.model.to(self.device)
    with torch.no_grad():
      for i, data in enumerate(dataset):
        input = data['input'].to(self.device)
        target = data['target'].to(self.device)
        input_mask = data['input_mask'].to(self.device)
        target_mask = data['target_mask'].to(self.device)

        generator_logit, loss = self.model.forward(input, target, input_mask, target_mask, labels=target)
        total_loss += loss.item()

    return total_loss / (len(dataset) - 1)

  def save(self, epoch, model, optimizer, losses, train_step):

    print(str(f'{self.checkpoint_path}/{self.model_name}.pth'))

    torch.save({
      'epoch': epoch,  # 현재 학습 epoch
      'model_state_dict': model.state_dict(),  # 모델 저장
      'optimizer_state_dict': optimizer.state_dict(),  # 옵티마이저 저장
      'losses': losses,  # Loss 저장
      'train_step': train_step,  # 현재 진행한 학습
    }, f'{self.checkpoint_path}/{self.model_name}.pth')



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import six

def keras_pad_fn(token_ids_batch, maxlen, pad_id=0, padding='post', truncating='post'):
    padded_token_ids_batch = pad_sequences(token_ids_batch,
                                            value=pad_id,  # vocab.transform_token2idx(PAD),
                                            padding=padding,
                                            truncating=truncating,
                                            maxlen=maxlen)
    return padded_token_ids_batch

# pad_sequences_fn in keras.preprocessing.sequence.pad_sequences
def pad_sequences(sequences, maxlen=None, dtype='int32',
                  padding='pre', truncating='pre', value=0.):
    """Pads sequences to the same length.

    This function transforms a list of
    `num_samples` sequences (lists of integers)
    into a 2D Numpy array of shape `(num_samples, num_timesteps)`.
    `num_timesteps` is either the `maxlen` argument if provided,
    or the length of the longest sequence otherwise.

    Sequences that are shorter than `num_timesteps`
    are padded with `value` at the end.

    Sequences longer than `num_timesteps` are truncated
    so that they fit the desired length.
    The position where padding or truncation happens is determined by
    the arguments `padding` and `truncating`, respectively.

    Pre-padding is the default.

    # Arguments
        sequences: List of lists, where each element is a sequence.
        maxlen: Int, maximum length of all sequences.
        dtype: Type of the output sequences.
            To pad sequences with variable length strings, you can use `object`.
        padding: String, 'pre' or 'post':
            pad either before or after each sequence.
        truncating: String, 'pre' or 'post':
            remove values from sequences larger than
            `maxlen`, either at the beginning or at the end of the sequences.
        value: Float or String, padding value.

    # Returns
        x: Numpy array with shape `(len(sequences), maxlen)`

    # Raises
        ValueError: In case of invalid values for `truncating` or `padding`,
            or in case of invalid shape for a `sequences` entry.
    """
    if not hasattr(sequences, '__len__'):
        raise ValueError('`sequences` must be iterable.')
    num_samples = len(sequences)

    lengths = []
    for x in sequences:
        try:
            lengths.append(len(x))
        except TypeError:
            raise ValueError('`sequences` must be a list of iterables. '
                             'Found non-iterable: ' + str(x))

    if maxlen is None:
        maxlen = np.max(lengths)

    # take the sample shape from the first non empty sequence
    # checking for consistency in the main loop below.
    sample_shape = tuple()
    for s in sequences:
        if len(s) > 0:
            sample_shape = np.asarray(s).shape[1:]
            break

    is_dtype_str = np.issubdtype(dtype, np.str_) or np.issubdtype(dtype, np.unicode_)
    if isinstance(value, six.string_types) and dtype != object and not is_dtype_str:
        raise ValueError("`dtype` {} is not compatible with `value`'s type: {}\n"
                         "You should set `dtype=object` for variable length strings."
                         .format(dtype, type(value)))

    x = np.full((num_samples, maxlen) + sample_shape, value, dtype=dtype)
    for idx, s in enumerate(sequences):
        if not len(s):
            continue  # empty list/array was found
        if truncating == 'pre':
            trunc = s[-maxlen:]
        elif truncating == 'post':
            trunc = s[:maxlen]
        else:
            raise ValueError('Truncating type "%s" '
                             'not understood' % truncating)

        # check `trunc` has expected shape
        trunc = np.asarray(trunc, dtype=dtype)
        if trunc.shape[1:] != sample_shape:
            raise ValueError('Shape of sample %s of sequence at position %s '
                             'is different from expected shape %s' %
                             (trunc.shape[1:], idx, sample_shape))

        if padding == 'post':
            x[idx, :len(trunc)] = trunc
        elif padding == 'pre':
            x[idx, -len(trunc):] = trunc
        else:
            raise ValueError('Padding type "%s" not understood' % padding)
    return x


if __name__ == '__main__':
    sequences = [[2, 4, 62], [2,35,12,24,2]]
    pad_res = pad_sequences(sequences, maxlen=10, dtype='int32', padding='pre', truncating='post', value=0.)
    keras_pad_res = keras_pad_fn(sequences, maxlen=10, pad_id=0, padding='post', truncating='post')
    print(pad_res)
    print(keras_pad_res)


import tensorflow as tf
from gensim.corpora import WikiCorpus
import os
import argparse

import ssl
ssl._create_default_https_context = ssl._create_unverified_context

def store(corpus, lang):
    base_path = os.getcwd()
    store_path = os.path.join(base_path, '{}_corpus'.format(lang))
    if not os.path.exists(store_path):
        os.mkdir(store_path)
    file_idx=1
    for text in corpus.get_texts():
        current_file_path = os.path.join(store_path, 'article_{}.txt'.format(file_idx))
        with open(current_file_path, 'w' , encoding='utf-8') as file:
            file.write(bytes(' '.join(text), 'utf-8').decode('utf-8'))
        #endwith
        file_idx += 1
    #endfor

def tokenizer_func(text: str, token_min_len: int, token_max_len: int, lower: bool) -> list:
    return [token for token in text.split() if token_min_len <= len(token) <= token_max_len]

def run(lang):
    origin='https://dumps.wikimedia.org/{}wiki/latest/{}wiki-latest-pages-articles.xml.bz2'.format(lang,lang)
    fname='{}wiki-latest-pages-articles.xml.bz2'.format(lang)
    file_path = tf.keras.utils.get_file(origin=origin, fname=fname, untar=False, extract=False)
    corpus = WikiCorpus(file_path, lemmatize=False, lower=False, tokenizer_func=tokenizer_func)
    store(corpus, lang)

if __name__ == '__main__':
    ARGS_PARSER = argparse.ArgumentParser()
    ARGS_PARSER.add_argument(
        '--lang',
        default='ko',
        type=str,
        help='language code to download from wikipedia corpus'
    )
    ARGS = ARGS_PARSER.parse_args()
    run(**vars(ARGS))


import re, collections

import os
from tokenizers.models import BPE
from tokenizers import Tokenizer
from tokenizers.decoders import ByteLevel as ByteLevelDecoder
from tokenizers.normalizers import NFKC, Sequence
from tokenizers.pre_tokenizers import ByteLevel
from tokenizers.trainers import BpeTrainer

class BPE_token(object):
    def __init__(self):
        self.tokenizer = Tokenizer(BPE())
        self.tokenizer.normalizer = Sequence([
            NFKC()
        ])
        self.tokenizer.pre_tokenizer = ByteLevel()
        self.tokenizer.decoder = ByteLevelDecoder()

    def bpe_train(self, paths):
        trainer = BpeTrainer(vocab_size=50000, show_progress=True, inital_alphabet=ByteLevel.alphabet(), special_tokens=[
            "<s>",
            "<pad>",
            "</s>",
            "<unk>",
            "<mask>"
        ])
        self.tokenizer.train(trainer, paths)

    def save_tokenizer(self, location, prefix=None):
        if not os.path.exists(location):
            os.makedirs(location)
        self.tokenizer.model.save(location, prefix)
        
def get_stats(dictionary):
    # 유니그램의 pair들의 빈도수를 카운트
    pairs = collections.defaultdict(int)
    for word, freq in dictionary.items():
        symbols = word.split()
        for i in range(len(symbols)-1):
            pairs[symbols[i],symbols[i+1]] += freq
    print('현재 pair들의 빈도수 :', dict(pairs))
    return pairs

def merge_dictionary(pair, v_in):
    v_out = {}
    bigram = re.escape(' '.join(pair))
    p = re.compile(r'(?<!\S)' + bigram + r'(?!\S)')
    for word in v_in:
        w_out = p.sub(''.join(pair), word)
        v_out[w_out] = v_in[word]
    return v_out

if __name__ == "__main__":
    num_merges = 10
    dictionary = {'l o w </w>' : 5,
             'l o w e r </w>' : 2,
             'n e w e s t </w>':6,
             'w i d e s t </w>':3
             }

    bpe_codes = {}
    bpe_codes_reverse = {}

    for i in range(num_merges):
        print("### Iteration {}".format(i + 1))
        pairs = get_stats(dictionary)
        best = max(pairs, key=pairs.get)
        dictionary = merge_dictionary(best, dictionary)

    bpe_codes[best] = i
    bpe_codes_reverse[best[0] + best[1]] = best

    print("new merge: {}".format(best))
    print("dictionary: {}".format(dictionary))
